"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.checkEnable = checkEnable;
exports.default = exports.cacheQueryEpic = exports.cacheSetEpic = exports.cacheClearEpic = void 0;

var _reduxPromiseEpic = require("../redux-promise-epic");

var _apiActionCreators = require("../sync/apiActionCreators");

var _cacheActionCreators = require("./cacheActionCreators");

var _util = require("./util");

/**
 * Provides a cache outside of Redux state that can optimistically update state
 * before an asynchronous API call returns
 *
 * @module CacheMiddleware
 */
function checkEnable() {
  if (typeof window !== 'undefined' && window.location) {
    var _ref = new URL(window.location.href),
        searchParams = _ref.searchParams;

    return !searchParams || !searchParams.has('__nocache');
  }

  return true;
}
/**
 * Listen for any action that should clear cached state
 *
 * Note that this will clear the cache without emitting an action
 */


var cacheClearEpic = function cacheClearEpic(cache) {
  return function (action) {
    if (action.type !== _cacheActionCreators.CACHE_CLEAR) {
      return Promise.resolve([]);
    }

    return cache.clear().then(function () {
      return [];
    });
  };
};

exports.cacheClearEpic = cacheClearEpic;

var getMemberId = function getMemberId(state) {
  return state.config.member.id.toString();
};
/**
 * Listen for any action that should set cached state with a
 * `{ queries, responses }` payload
 *
 * API_RESP_SUCCESS means there is fresh data ready to be stored - extract the
 * queries and their responses, then dispatch `CACHE_SET` actions with each
 * pair
 *
 * Not that this will set the cache without emitting an action
 */


var cacheSetEpic = function cacheSetEpic(cache) {
  return function (action, store) {
    if (![_apiActionCreators.API_RESP_SUCCESS, _cacheActionCreators.CACHE_SET].some(function (type) {
      return action.type === type;
    })) {
      return Promise.resolve([]);
    }

    var _action$payload = action.payload,
        query = _action$payload.query,
        response = _action$payload.response;
    var writeCache = (0, _util.cacheWriter)(cache, getMemberId(store.getState()));
    return writeCache(query, response).then(function () {
      return [];
    });
  };
};
/**
 * Listen for any action that should query the cache using a payload of queries
 *
 * Promises are heavily used in CACHE_REQUEST because each query results in
 * an async 'get' (Promise) from the Cache - all 'gets' happen in parallel and
 * the results are collated into a single response object containing the cache
 * hits.
 */


exports.cacheSetEpic = cacheSetEpic;

var cacheQueryEpic = function cacheQueryEpic(cache) {
  return function (action, store) {
    if (action.type !== _apiActionCreators.API_REQ) {
      return Promise.resolve([]);
    }

    var queries = action.payload;
    var readCache = (0, _util.cacheReader)(cache, getMemberId(store.getState()));
    return Promise.all(queries.map(readCache)) // read cache for all queries
    .then(function (responses) {
      return responses.filter(function (_ref2) {
        var query = _ref2.query,
            response = _ref2.response;
        return response;
      });
    }) // filter out the misses
    .then(function (hits) {
      return hits.map(_cacheActionCreators.cacheSuccess);
    }); // map the hits onto cacheSuccess actions
  };
};

exports.cacheQueryEpic = cacheQueryEpic;

var getCacheEpic = function getCacheEpic() {
  var cache = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : (0, _util.makeCache)();
  return checkEnable() ? (0, _reduxPromiseEpic.combineEpics)(cacheClearEpic(cache), cacheSetEpic(cache), cacheQueryEpic(cache)) : function (action) {
    return Promise.resolve([]);
  };
};

var _default = getCacheEpic;
exports.default = _default;
//# sourceMappingURL=index.js.map