"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.makeCache = makeCache;
exports.cacheWriter = exports.cacheReader = void 0;

/**
 * This function performs feature sniffing to determine whether the preferred
 * IndexedDB cache is available, otherwise it falls back to a simple
 * plain-object-based cache that will only survive as long as the request.
 *
 * The cache object methods are thin wrappers around their IndexedDB
 * ObjectStore equivalents
 *
 * {@link https://developer.mozilla.org/en-US/docs/Web/API/IDBObjectStore}
 */
function makeCache() {
  if (typeof window === 'undefined' || !window.indexedDB) {
    var _data = {};
    return {
      get: function get(key) {
        return Promise.resolve(_data[key]);
      },
      set: function set(key, val) {
        _data[key] = val;
        return Promise.resolve(true);
      },
      delete: function _delete(key) {
        delete _data[key];
        return Promise.resolve(true);
      },
      clear: function clear() {
        Object.keys(_data).forEach(function (key) {
          return delete _data[key];
        });
        return Promise.resolve(true);
      }
    };
  }

  return require('idb-keyval');
}

var makeKey = function makeKey(memberId, query) {
  return "".concat(memberId).concat(JSON.stringify(query));
};
/**
 * Generates a function that can read queries and return hits in the supplied cache
 */


var cacheReader = function cacheReader(cache, memberId) {
  return function (query) {
    return cache.get(makeKey(memberId, query)).then(function (response) {
      return {
        query: query,
        response: response
      };
    }).catch(function (err) {
      return {
        query: query,
        response: null
      };
    });
  };
}; // errors don't matter - just return null

/**
 * Generates a function that can write query-response values into cache
 *
 * It will ignore non-GET responses and any responses to queries that have
 * opted-out of caching with `query.meta.noCache`
 */


exports.cacheReader = cacheReader;

var cacheWriter = function cacheWriter(cache, memberId) {
  return function (query, response) {
    var method = (query.meta || {}).method || 'get';

    if (method.toLowerCase() !== 'get' || query.meta && query.meta.noCache) {
      // skip cache writing
      return Promise.resolve(true);
    }

    return cache.set(makeKey(memberId, query), response);
  };
};

exports.cacheWriter = cacheWriter;
//# sourceMappingURL=util.js.map