"use strict";

var _util = require("./util");

describe('cache utils', function () {
  it('creates a Promise-based cache', function () {
    var cache = (0, _util.makeCache)();
    expect(cache.get).toEqual(jasmine.any(Function));
    expect(cache.get()).toEqual(jasmine.any(Promise));
    expect(cache.set).toEqual(jasmine.any(Function));
    expect(cache.set()).toEqual(jasmine.any(Promise));
    expect(cache.delete).toEqual(jasmine.any(Function));
    expect(cache.delete()).toEqual(jasmine.any(Promise));
  });
  it('sets and gets from cache', function () {
    var cache = (0, _util.makeCache)();
    return cache.set('foo', 'bar').then(function () {
      return cache.get('foo');
    }).then(function (value) {
      return expect(value).toEqual('bar');
    }).then(function () {
      return cache.set('foo', 'baz');
    }).then(function () {
      return cache.get('foo');
    }).then(function (value) {
      return expect(value).toEqual('baz');
    });
  });
  it('deletes from cache', function () {
    var cache = (0, _util.makeCache)();
    return cache.set('foo', 'bar').then(function () {
      return cache.delete('foo');
    }).then(function () {
      return cache.get('foo');
    }).then(function (value) {
      return expect(value).toBeUndefined();
    }).catch(function (err) {
      return expect(err).toEqual(jasmine.any(Error));
    });
  });
  it('cacheWriter writes to cache', function () {
    var cache = (0, _util.makeCache)();
    var query = {
      foo: 'baz'
    };
    var response = 'bar';
    return (0, _util.cacheWriter)(cache, 'memberId')(query, response).then(function () {
      return cache.get("memberId".concat(JSON.stringify(query)));
    }).then(function (cachedResponse) {
      return expect(cachedResponse).toEqual(response);
    });
  });
  it('cacheWriter does not write non-GET queries to cache', function () {
    var cache = (0, _util.makeCache)();
    var query = {
      foo: 'baz',
      meta: {
        method: 'post'
      }
    };
    var response = 'bar';
    return (0, _util.cacheWriter)(cache)(query, response).then(function () {
      return cache.get(JSON.stringify(query));
    }).then(function (cachedResponse) {
      return expect(cachedResponse).toEqual(undefined);
    });
  });
  it('cacheWriter does not write `noCache` queries to cache', function () {
    var cache = (0, _util.makeCache)();
    var query = {
      foo: 'baz',
      meta: {
        noCache: true
      }
    };
    var response = 'bar';
    return (0, _util.cacheWriter)(cache)(query, response).then(function () {
      return cache.get(JSON.stringify(query));
    }).then(function (cachedResponse) {
      return expect(cachedResponse).toEqual(undefined);
    });
  });
  it('cacheReader reads from cache and returns result with query', function () {
    var cache = (0, _util.makeCache)();
    var requestCache = (0, _util.cacheReader)(cache);
    var query = {
      foo: 'baz'
    };
    var nonCachedQuery = {
      bing: 'bong'
    };
    var response = 'bar';
    return (0, _util.cacheWriter)(cache)(query, response).then(function () {
      return requestCache(query);
    }).then(function (_ref) {
      var cachedQuery = _ref.query,
          cachedResponse = _ref.response;
      // expecting two-element array response
      expect(cachedQuery).toEqual(query);
      expect(cachedResponse).toEqual(response);
    }).then(function () {
      return requestCache(nonCachedQuery);
    }) // test with un-cached query
    .then(function (_ref2) {
      var cachedQuery = _ref2.query,
          cachedResponse = _ref2.response;
      expect(cachedQuery).toEqual(nonCachedQuery);
      expect(cachedResponse).toBeUndefined();
    });
  });
});
//# sourceMappingURL=util.test.js.map